#include <sys/types.h>
#include <sys/time.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/ioctl.h>
#include <sys/syslog.h>
#include <fcntl.h>
#include <linux/pcmcia.h>

void error(char *fmt, ...)
  {
  va_list args;
  va_start(args, fmt);
  vsyslog(LOG_ERR, fmt, args);
  va_end(args);
  syslog(LOG_ERR, "exiting");
  exit(EXIT_FAILURE);
  }

void info(char *fmt, ...)
  {
  va_list args;
  va_start(args, fmt);
  vsyslog(LOG_INFO, fmt, args);
  va_end(args);
  }
  
void enable(int fd, char *name)
  {
  int val;
  struct pcmcia_io_map io[2] = {
    { 0, MAP_ACTIVE, 0x300, 0x30F },
    { 1, MAP_ACTIVE|MAP_CS16|MAP_16BIT, 0x310, 0x31F }
    };
  struct pcmcia_mem_map mem = {
    0, MAP_ACTIVE|MAP_ATTRIB, 0, 0xD6000L, 0xD6FFFL, 0x2000000L
    };
  struct pcmcia_register cor = { PCMCIA_COR, 0x41, 0x0008 };
  
  val = PC_VCC_5V | PC_RESET;
  if (ioctl(fd, PCMCIA_SET_FLAGS, &val) < 0)
    error("ioctl(%s, PCMCIA_SET_FLAGS): %m", name);
  usleep(100000); val = PC_VCC_5V;
  if (ioctl(fd, PCMCIA_SET_FLAGS, &val) < 0)
    error("ioctl(%s, PCMCIA_SET_FLAGS): %m", name);
  for (;;) {
    if (ioctl(fd, PCMCIA_GET_STATUS, &val) < 0)
      error("ioctl(%s, PCMCIA_GET_STATUS): %m", name);
    if (val & PC_READY) break;
    printf("."); fflush(stdout);
    usleep(100000);
    }
  val = PC_VCC_5V | PC_IOCARD;
  if (ioctl(fd, PCMCIA_SET_FLAGS, &val) < 0)
    error("ioctl(%s, PCMCIA_SET_FLAGS): %m", name);
  if (ioctl(fd, PCMCIA_SET_IO_MAP, &io[0]) < 0)
    error("ioctl(%s, PCMCIA_SET_IO_MAP): %m", name);
  if (ioctl(fd, PCMCIA_SET_IO_MAP, &io[1]) < 0)
    error("ioctl(%s, PCMCIA_SET_IO_MAP): %m", name);
  if (ioctl(fd, PCMCIA_SET_MEM_MAP, &mem) < 0)
    error("ioctl(%s, PCMCIA_SET_MEM_MAP): %m", name);
  val = 5;
  if (ioctl(fd, PCMCIA_SET_IRQ, &val) < 0)
    error("ioctl(%s, PCMCIA_SET_IRQ): %m", name);
  /* Set configuration option register */
  if (ioctl(fd, PCMCIA_SET_REG, &cor) < 0)
    error("ioctl(%s, PCMCIA_SET_REG): %m", name);
  if (system("/etc/rc.d/pc_eth0 start") != 0)
    error("system(pc_eth0 stop): %m");
  }
  
void disable(int fd, char *name)
  {
  int val;
  val = PC_PWR_OFF;
  if (ioctl(fd, PCMCIA_SET_FLAGS, &val) < 0)
    error("ioctl(%s, PCMCIA_SET_FLAGS): %m", name);
  if (system("/etc/rc.d/pc_eth0 stop") != 0)
    error("system(pc_eth0 stop): %m");
  }
  
void main(int argc, char *argv[])
  {
  int fd[8], i, ret, val, event;
  int handsoff = 0;
  fd_set fds;

  openlog("dumbd", LOG_PID, LOG_DAEMON);
  
  argv++; argc--;
  if (argc > 0)
    if (strcmp(*argv, "-h") == 0) {
      handsoff = 1;
      argv++; argc--;
      }
  if (argc < 1) error("no devices to watch");
  info("starting");
  
  for (i = 0; i < argc; i++) {
    if ((fd[i] = open(argv[i], O_RDWR)) < 0)
      error("open(%s): %m", argv[i]);
    }

  for (i = 0; i < argc; i++) {
    if (ioctl(fd[i], PCMCIA_GET_STATUS, &val) < 0)
      error("ioctl(%s, PCMCIA_GET_STATUS): %m", argv[i]);
    if (val & PC_DETECT) {
      info("%s: card present", argv[i]);
      if (!handsoff) enable(fd[i], argv[i]);
      }
    }

  /* Now become a daemon */
  ret = fork();
  if (ret > 0) exit(0);
  if (ret == -1) error("fork(): %m");

  for (;;) {
    
    FD_ZERO(&fds);
    for (i = 0; i < argc; i++) FD_SET(fd[i], &fds);
    if (select(10, &fds, NULL, NULL, NULL) < 0)
      error("select(): %m");
    sleep(1);
    
    for (i = 0; i < argc; i++) {
      
      if (!FD_ISSET(fd[i], &fds)) continue;
      if ((ret = read(fd[i], &event, 4)) == -1) {
	if (errno != EAGAIN)
	  error("read(%s) failed: %m\n", argv[i]);
	}
      
      if ((ret == 4) && (event & PC_DETECT)) {
	if (ioctl(fd[i], PCMCIA_GET_STATUS, &val) < 0)
	  error("ioctl(%s, PCMCIA_GET_STATUS): %m", argv[i]);
	if (val & PC_DETECT) {
	  info("%s: card inserted", argv[i]);
	  enable(fd[i], argv[i]);
	  }
	else {
	  info("%s: card removed", argv[i]);
	  disable(fd[i], argv[i]);
	  }
	
	}
      }
    
    }
  }
