/*======================================================================

    Device driver for Intel 82365 PCMCIA controller

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <sys/types.h>
#include <linux/fcntl.h>
#include <linux/string.h>

#include <asm/io.h>
#include <asm/bitops.h>
#include <asm/segment.h>
#include <asm/system.h>

#ifdef USER_DRV
#include "user_drv.h"
#else
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#endif

#ifndef PHYS_ADDR
#define PHYS_ADDR 0xd0000
#endif

/* sigh */
static u_long phys_addr;
extern int pc_debug;

#include <linux/pcmcia.h>
#include "i82365.h"

/*====================================================================*/

static u_char i365_get(u_short slot, u_short reg)
  {
  u_short port = I365_PORT(slot);
  u_char val = I365_REG(slot, reg);
  cli();
  outb_p(val, port); val = inb_p(port+1);
  sti();
  return val;
  }

static void i365_set(u_short slot, u_short reg, u_char data)
  {
  u_short port = I365_PORT(slot);
  u_char val = I365_REG(slot, reg);
  cli();
  outb_p(val, port); outb_p(data, port+1);
  sti();
  }

static void i365_bset(u_short slot, u_short reg, u_char mask)
  {
  u_char d = i365_get(slot, reg);
  d |= mask;
  i365_set(slot, reg, d);
  }

static void i365_bclr(u_short slot, u_short reg, u_char mask)
  {
  u_char d = i365_get(slot, reg);
  d &= ~mask;
  i365_set(slot, reg, d);
  }

static u_short i365_get_pair(u_short slot, u_short reg)
  {
  u_short a, b;
  a = i365_get(slot, reg);
  b = i365_get(slot, reg+1);
  return (a + (b<<8));
  }

static void i365_set_pair(u_short slot, u_short reg, u_short data)
  {
  i365_set(slot, reg, data & 0xFF);
  i365_set(slot, reg+1, data >> 8);
  }

/*====================================================================*/

void i365_csc_on(u_short slot, int irq);
void i365_csc_off(u_short slot);
int i365_csc_poll(u_short slot);

int i365_ioctl(u_short slot, u_int cmd, u_long arg);

int i365_probe(struct pcmcia_ctrl *ctrl)
  {
  int irqlist[] = { 15, 14, 12, 11, 10, 9, 7, 5, 4, 3, 0 };
  int val, id, slot, *irqp;

  printk("Intel 82365 PCMCIA probe: ");
  id = slot = 0;
  if (check_region(I365_BASE, 4) == 0)
    for (; slot < 8; slot++) {
      val = i365_get(slot, I365_IDENT);
      if ((val != 0x82) && (val != 0x83)) break;
      id = val;
      }
  
  if (slot == 0) {
    printk("not found.\n");
    return -ENODEV;
    }

  snarf_region(I365_BASE, 4);
  if (id > 0x82) {
    for (irqp = irqlist; *irqp != 0; irqp++) {
      if (request_irq(*irqp, NULL) == -EBUSY) continue;
	autoirq_setup(0);
	i365_csc_on(0, *irqp);
	i365_bset(0, I365_GENCTL, I365_CTL_SW_IRQ);
        i365_csc_off(0);
        if (*irqp == autoirq_report(0)) break;
      }
    ctrl->csc_irq = *irqp;
    }
  else
    ctrl->csc_irq = 0;
  printk("%d slots, version 0x%2X, ", slot, id);
  if (ctrl->csc_irq == 0)
    printk("polled card status mode\n");
  else
    printk("status change on IRQ %d\n", ctrl->csc_irq);
  ctrl->csc_on = i365_csc_on;
  ctrl->csc_off = i365_csc_off;
  ctrl->csc_poll = i365_csc_poll;
  ctrl->ioctl = i365_ioctl;
  phys_addr = PHYS_ADDR;
  return slot;
  } /* i365_probe */
  
/*====================================================================*/

void i365_csc_on(u_short slot, int irq)
  {
  /* Turn on all card status change interrupts */
  i365_set(slot, I365_CSCINT, (irq << 4) | I365_CSC_ANY);
  i365_bclr(slot, I365_INTCTL, I365_INTR_ENA);
  i365_bset(slot, I365_GENCTL, I365_CTL_GPI_ENA);
  /* flush current card status changes */
  i365_get(slot, I365_CSC);
  } /* i365_csc_on */

void i365_csc_off(u_short slot)
  {
  /* Turn off all card status change interrupts */
  i365_set(slot, I365_CSCINT, 0x00);
  i365_bclr(slot, I365_GENCTL, I365_CTL_GPI_ENA);
  } /* i365_csc_off */

int i365_csc_poll(u_short slot)
  {
  int ret;
  /* Anything to report? */
  int csc = i365_get(slot, I365_CSC);
  if (i365_get(slot, I365_INTCTL) & I365_PC_IOCARD)
    ret  = (csc & I365_CSC_STSCHG) ? PC_STSCHG : 0;
  else {
    ret  = (csc & I365_CSC_BVD1) ? PC_BATDEAD : 0;
    ret |= (csc & I365_CSC_BVD2) ? PC_BATWARN : 0;
    }
  ret |= (csc & I365_CSC_READY) ? PC_READY : 0;
  ret |= (csc & I365_CSC_DETECT) ? PC_DETECT : 0;
  ret |= (csc & I365_CSC_GPI) ? PC_GPI : 0;
  return ret;
  } /* i365_csc_poll */

/*====================================================================*/

static void delay(int len)
  {
  int start = jiffies;
  int boguscount = 150000;
  sti();
  while (jiffies - start < len)
    if (boguscount-- < 0) {
      printk("jiffy failure (t=%ld)...", jiffies);
      break;
      }
  } /* delay */
  
/*====================================================================*/

int i365_get_status(u_short slot, u_int *value)
  {
  u_char status;
  status = i365_get(slot, I365_STATUS);
  *value = ((status & I365_CS_DETECT) == I365_CS_DETECT) ? PC_DETECT : 0;
  if (i365_get(slot, I365_INTCTL) & I365_PC_IOCARD)
    *value |= (status & I365_CS_STSCHG) ? 0 : PC_STSCHG;
  else {
    *value |= (status & I365_CS_BVD1) ? 0 : PC_BATDEAD;
    *value |= (status & I365_CS_BVD2) ? 0 : PC_BATWARN;
    }
  *value |= (status & I365_CS_WRPROT) ? PC_WRPROT : 0;
  *value |= (status & I365_CS_READY) ? PC_READY : 0;
  *value |= (status & I365_CS_POWERON) ? PC_POWERON : 0;
  *value |= (status & I365_CS_GPI) ? 0 : PC_GPI;
  return 0;
  } /* i365_get_status */
  
/*====================================================================*/

int i365_get_flags(u_short slot, u_int *value)
  {
  u_char reg, vcc, vpp;
  reg = i365_get(slot, I365_POWER);
  *value  = (reg & I365_PWR_AUTO) ? PC_PWR_AUTO : 0;
  vcc = reg & I365_VCC_MASK; vpp = reg & I365_VPP_MASK;
  *value |= (vcc == I365_VCC_3V) ? PC_VCC_3V : 0;
  *value |= (vcc == I365_VCC_5V) ? PC_VCC_5V : 0;
  *value |= (vpp == I365_VPP_5V) ? PC_VPP_5V : 0;
  *value |= (vpp == I365_VPP_12V) ? PC_VPP_12V : 0;
  reg = i365_get(slot, I365_INTCTL);
  *value |= (reg & I365_PC_IOCARD) ? PC_IOCARD : 0;
  *value |= (reg & I365_PC_RESET) ? 0 : PC_RESET;
  return 0;
  } /* i365_get_flags */
  
/*====================================================================*/

int i365_set_flags(u_short slot, u_int *value)
  {
  u_char reg = i365_get(slot, I365_POWER);
  reg &= ~(I365_PWR_AUTO | I365_VCC_MASK | I365_VPP_MASK);
  reg |= (*value & PC_PWR_AUTO) ? I365_PWR_AUTO : 0;
  reg |= (*value & PC_VCC_3V) ? I365_VCC_3V|I365_PWR_OUT : 0;
  reg |= (*value & PC_VCC_5V) ? I365_VCC_5V|I365_PWR_OUT : 0;
  reg |= (*value & PC_VPP_5V) ? I365_VPP_5V : 0;
  reg |= (*value & PC_VPP_12V) ? I365_VPP_12V : 0;
  i365_set(slot, I365_POWER, reg);
  delay(2);
  reg = i365_get(slot, I365_INTCTL);
  reg &= ~(I365_PC_IOCARD | I365_PC_RESET);
  /* Note that the reset signal is inverted */
  reg |= (*value & PC_RESET) ? 0 : I365_PC_RESET;
  reg |= (*value & PC_IOCARD) ? I365_PC_IOCARD : 0;
  i365_set(slot, I365_INTCTL, reg);
  return 0;
  } /* i365_set_flags */
  
/*====================================================================*/

int i365_get_irq(u_short slot, u_int *value)
  {
  u_char ctl = i365_get(slot, I365_INTCTL);
  *value = ctl & 0x0F;
  return 0;
  } /* i365_get_irq */
  
/*====================================================================*/

int i365_set_irq(u_short slot, u_int *value)
  {
  u_char ctl = i365_get(slot, I365_INTCTL);
  if (*value > 15) return -EINVAL;
  ctl = (ctl & 0xf0) | *value;
  i365_set(slot, I365_INTCTL, ctl);
  return 0;
  } /* i365_set_irq */
  
/*====================================================================*/

int i365_get_io_map(u_short slot, struct pcmcia_io_map *io)
  {
  u_char map, ioctl, addr;
  map = io->map;
  if (map > 2) return -EINVAL;
  io->start = i365_get_pair(slot, I365_IO(map)+I365_W_START);
  io->stop = i365_get_pair(slot, I365_IO(map)+I365_W_STOP);
  ioctl = i365_get(slot, I365_IOCTL);
  addr = i365_get(slot, I365_ADDRWIN);
  io->flags  = (addr & I365_ENA_IO(map)) ? MAP_ACTIVE : 0;
  io->flags |= (ioctl & I365_IOCTL_WAIT(map)) ? MAP_WAIT : 0; 
  io->flags |= (ioctl & I365_IOCTL_0WS(map)) ? MAP_0WS : 0;
  io->flags |= (ioctl & I365_IOCTL_CS16(map)) ? MAP_CS16 : 0;
  io->flags |= (ioctl & I365_IOCTL_16BIT(map)) ? MAP_16BIT : 0;
  return 0;
  } /* i365_get_io_map */

/*====================================================================*/

int i365_set_io_map(u_short slot, struct pcmcia_io_map *io)
  {
  u_char map, ioctl;
  map = io->map;
  if ((map > 2) || (io->start > 0xFFFF) || (io->stop > 0xFFFF) ||
      (io->stop < io->start)) return -EINVAL;
  /* Turn off the window before changing anything */
  if (i365_get(slot, I365_ADDRWIN) & I365_ENA_IO(map))
    i365_bclr(slot, I365_ADDRWIN, I365_ENA_IO(map));
  i365_set_pair(slot, I365_IO(map)+I365_W_START, io->start);
  i365_set_pair(slot, I365_IO(map)+I365_W_STOP, io->stop);
  ioctl = i365_get(slot, I365_IOCTL) & ~I365_IOCTL_MASK(map);
  if (io->flags & MAP_WAIT) ioctl |= I365_IOCTL_WAIT(map);
  if (io->flags & MAP_0WS) ioctl |= I365_IOCTL_0WS(map);
  if (io->flags & MAP_CS16) ioctl |= I365_IOCTL_CS16(map);
  if (io->flags & MAP_16BIT) ioctl |= I365_IOCTL_16BIT(map);
  i365_set(slot, I365_IOCTL, ioctl);
  /* Turn on the window if necessary */
  if (io->flags & MAP_ACTIVE)
    i365_bset(slot, I365_ADDRWIN, I365_ENA_IO(map));
  return 0;
  } /* i365_set_io_map */

/*====================================================================*/

int i365_get_mem_map(u_short slot, struct pcmcia_mem_map *mem)
  {
  u_short base, i;
  u_char map, addr;

  map = mem->map;
  if (map > 4) return -EINVAL;
  addr = i365_get(slot, I365_ADDRWIN);
  mem->flags = (addr & I365_ENA_MEM(map)) ? MAP_ACTIVE : 0;
  base = I365_MEM(map);
  
  i = i365_get_pair(slot, base+I365_W_START);
  mem->flags |= (i & I365_MEM_16BIT) ? MAP_16BIT : 0;
  mem->flags |= (i & I365_MEM_0WS) ? MAP_0WS : 0;
  mem->sys_start = ((long)(i & 0x0FFF) << 12);
  
  i = i365_get_pair(slot, base+I365_W_STOP);
  mem->extra_ws  = (i & I365_MEM_WS0) ? 1 : 0;
  mem->extra_ws += (i & I365_MEM_WS1) ? 2 : 0;
  mem->sys_stop = ((long)(i & 0x0FFF) << 12) + 0x0FFF;
  
  i = i365_get_pair(slot, base+I365_W_OFF);
  mem->flags |= (i & I365_MEM_WRPROT) ? MAP_WRPROT : 0;
  mem->flags |= (i & I365_MEM_REG) ? MAP_ATTRIB : 0;
  mem->card_start = ((long)(i & 0x3FFF) << 12) + mem->sys_start;
  mem->card_start &= 0x3FFFFFF;
  return 0;
  } /* i365_get_mem_map */

/*====================================================================*/
  
int i365_set_mem_map(u_short slot, struct pcmcia_mem_map *mem)
  {
  u_short base, i;
  u_char map;

  map = mem->map;
  if ((map > 4) || (mem->card_start > 0x3FFFFFF) ||
      (mem->sys_start > 0xFFFFFF) || (mem->sys_stop > 0xFFFFFF) ||
      (mem->sys_start > mem->sys_stop) || (mem->extra_ws > 3))
    return -EINVAL;
  /* Turn off the window before changing anything */
  if (i365_get(slot, I365_ADDRWIN) & I365_ENA_MEM(map))
    i365_bclr(slot, I365_ADDRWIN, I365_ENA_MEM(map));
  base = I365_MEM(map);

  i = mem->sys_start >> 12;
  if (mem->flags & MAP_16BIT) i |= I365_MEM_16BIT;
  if (mem->flags & MAP_0WS) i |= I365_MEM_0WS;
  i365_set_pair(slot, base+I365_W_START, i);
  
  i = mem->sys_stop >> 12;
  if (mem->extra_ws & 1) i |= I365_MEM_WS0;
  if (mem->extra_ws & 2) i |= I365_MEM_WS1;
  i365_set_pair(slot, base+I365_W_STOP, i);
  
  i = ((mem->card_start - mem->sys_start) >> 12) & 0x3FFF;
  if (mem->flags & MAP_WRPROT) i |= I365_MEM_WRPROT;
  if (mem->flags & MAP_ATTRIB) i |= I365_MEM_REG;
  i365_set_pair(slot, base+I365_W_OFF, i);
                                             
  /* Turn on the window if necessary */
  if (mem->flags & MAP_ACTIVE)
    i365_bset(slot, I365_ADDRWIN, I365_ENA_MEM(map));
  return 0;
  } /* i365_set_mem_map */

/*====================================================================*/

int i365_map_attr(int slot, struct pcmcia_mem_map *old,
		  struct pcmcia_mem_map *tmp)
  {
  int err;
  
  old->map = 4;
  if ((err = i365_get_mem_map(slot, old)) < 0)
    return err;
  
  tmp->map = 4;
  tmp->sys_start = phys_addr;
  tmp->sys_stop = phys_addr + sizeof(struct pcmcia_attr_mem);
  tmp->card_start = 0;
  tmp->extra_ws = 0;
  tmp->flags = MAP_ACTIVE | MAP_ATTRIB;
  if ((err = i365_set_mem_map(slot, tmp)) < 0)
    return err;
#ifdef USER_DRV
  tmp->sys_start = (u_long)mmap_addr;
#endif
  return 0;
  }

int i365_get_attr_mem(u_short slot, struct pcmcia_attr_mem *attrmem)
  {
  struct pcmcia_mem_map old, tmp;
  int err;
  
  if ((err = i365_map_attr(slot, &old, &tmp)) < 0)
    return err;
  
  memcpy((char *) attrmem->buf, (char *) tmp.sys_start,
	 sizeof(attrmem->buf));
     
  if ((err = i365_set_mem_map(slot, &old)) < 0)
    return err;
  
  return 0;
  }
   
/*====================================================================*/

int i365_getset_register(u_short slot, struct pcmcia_register *reg,
			 int get)
  {
  struct pcmcia_mem_map old, tmp;
  int err;

  /* Hack!  This function is intended to get and set PCMCIA */
  /* registers.  However, if reg & 0x80, reg & 0x7f is a 365SL */
  /* register number and reg->val is (should be) its value. */
  if (reg->reg & 0x80) {
    reg->reg &= 0x7f;
    if (get)
      reg->val = i365_get(slot, reg->reg);
    else
      i365_set(slot, reg->reg, reg->val);
    if  (pc_debug)
      printk("i365_%s_register: i365 reg %#x = %#x\n", get ?
	     "get" : "set", reg->reg, reg->val);
    return 0;
    }
  
  if ((err = i365_map_attr(slot, &old, &tmp)) < 0) {
    if (pc_debug)
      printk("i365_%s_register: map_attr = %d\n", get ? "get"
	     : "set", err);
    return err;
    }
  
  if (get)
    memcpy((char *) &reg->val,
	   (char *) (tmp.sys_start + reg->ofs + reg->reg),
	   sizeof(reg->val));
  else
    memcpy((char *) (tmp.sys_start + reg->ofs + reg->reg),
	   (char *) &reg->val, sizeof(reg->val));
     
  if ((err = i365_set_mem_map(slot, &old)) < 0) {
    if (pc_debug)
      printk("i365_%s_register: set_mem_map = %d\n", get ? "get"
	     : "set", err);
    return err;
    }
  
  return 0;
  }

/*====================================================================*/

int i365_ioctl(u_short slot, u_int cmd, u_long arg)
  {
  int err;
  u_int value;
  struct pcmcia_io_map iomap;
  struct pcmcia_mem_map memmap;
  struct pcmcia_attr_mem attrmem;
  struct pcmcia_register reg;

  err = 0;
  switch (cmd) {
    case PCMCIA_GET_STATUS:
      err = i365_get_status(slot, &value);
      memcpy_tofs((char *)arg, &value, sizeof(int));
      break;
    case PCMCIA_GET_FLAGS:
      err = i365_get_flags(slot, &value);
      memcpy_tofs((char *)arg, &value, sizeof(int));
      break;
    case PCMCIA_SET_FLAGS:
      memcpy_fromfs(&value, (char *)arg, sizeof(int));
      err = i365_set_flags(slot, &value);
      break;
    case PCMCIA_GET_IRQ:
      err = i365_get_irq(slot, &value);
      memcpy_tofs((char *)arg, &value, sizeof(int));
      break;
    case PCMCIA_SET_IRQ:
      memcpy_fromfs(&value, (char *)arg, sizeof(int));
      err = i365_set_irq(slot, &value);
      break;
    case PCMCIA_GET_IO_MAP:
      memcpy_fromfs(&iomap, (char *)arg, sizeof(struct pcmcia_io_map));
      err = i365_get_io_map(slot, &iomap);
      memcpy_tofs((char *)arg, &iomap, sizeof(struct pcmcia_io_map));
      break;
    case PCMCIA_SET_IO_MAP:
      memcpy_fromfs(&iomap, (char *)arg, sizeof(struct pcmcia_io_map));
      err = i365_set_io_map(slot, &iomap);
      break;
    case PCMCIA_GET_MEM_MAP:
      memcpy_fromfs(&memmap, (char *)arg, sizeof(struct pcmcia_mem_map));
      err = i365_get_mem_map(slot, &memmap);
      memcpy_tofs((char *)arg, &memmap, sizeof(struct pcmcia_mem_map));
      break;
    case PCMCIA_SET_MEM_MAP:
      memcpy_fromfs(&memmap, (char *)arg, sizeof(struct pcmcia_mem_map));
      err = i365_set_mem_map(slot, &memmap);
      break;
    case PCMCIA_GET_ATTR_MEM:
      err = i365_get_attr_mem(slot, &attrmem);
      memcpy_tofs((char *)arg, &attrmem, sizeof(struct pcmcia_attr_mem));
      break;
    case PCMCIA_GET_REG:
      memcpy_fromfs(&reg, (char *)arg, sizeof(struct pcmcia_register));
      err = i365_getset_register(slot, &reg, 1);
      memcpy_tofs((char *) arg, &reg, sizeof(struct pcmcia_register));
      break;
    case PCMCIA_SET_REG:
      memcpy_fromfs(&reg, (char *)arg, sizeof(struct pcmcia_register));
      err = i365_getset_register(slot, &reg, 0);
      break;
    default:
      err = -EINVAL;
      if (pc_debug)
	printk("i365_ioctl(%d, %x, %lx), cmd unknown\n", slot, cmd, arg);
      break;
    }
  if (err && pc_debug) 
       printk("i365_ioctl(%d, %x, %lx) = %d\n", slot, cmd, arg, err);
  return err;
  } /* i365_ioctl */
