/*======================================================================

    High-level PCMCIA driver interface

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <sys/types.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/config.h>

#include <asm/io.h>
#include <asm/bitops.h>
#include <asm/segment.h>
#include <asm/system.h>

#ifdef USER_DRV
#include "user_drv.h"
#else
#include <linux/fcntl.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#endif

#include <linux/pcmcia.h>
#include "i82365.h"

#ifdef PCMCIA_DEBUG
int pc_debug = PCMCIA_DEBUG;
#else
int pc_debug = 0;
#endif

/*====================================================================*/

#define MAX_SLOTS 8

static u_short slots_present = 0;

static struct pcmcia_slot slot_table[MAX_SLOTS] = {
  { 0, 0, 0 },
  { 0, 0, 0 },
  { 0, 0, 0 },
  { 0, 0, 0 },
  { 0, 0, 0 },
  { 0, 0, 0 },
  { 0, 0, 0 },
  { 0, 0, 0 }
  };

static struct pcmcia_ctrl ctrl;

/*====================================================================*/

static int pcmcia_read(struct inode *inode, struct file *file,
		       char *buf, int count);
static int pcmcia_select(struct inode *inode, struct file *file,
                         int sel_type, select_table *wait);
static int pcmcia_ioctl(struct inode * inode, struct file * file,
                        u_int cmd, u_long arg);
static int pcmcia_open(struct inode * inode, struct file * file);
static void pcmcia_release(struct inode * inode, struct file * file);
static void pcmcia_interrupt(int irq);

static struct file_operations pcmcia_fops = {
  NULL,           /* lseek */
  pcmcia_read,    /* read */
  NULL,           /* write */
  NULL,           /* readdir */
  pcmcia_select,  /* select */
  pcmcia_ioctl,   /* ioctl */
  NULL,           /* mmap */
  pcmcia_open,    /* open */
  pcmcia_release, /* release */
  NULL            /* fsync */
  };

/*====================================================================*/

u_long pcmcia_init(u_long kmem_start, u_long kmem_end)
  {
  int ret;
  
#ifdef CONFIG_I82365
  ret = i365_probe(&ctrl);
  if (ret > 0) slots_present = ret;
#endif
  
#ifdef CONFIG_TCIC
  ret = tcic_probe(&ctrl);
  if (ret > 0) slots_present = ret;
#endif

  /* Grab an interrupt */
  if ((slots_present > 0) && (ctrl.csc_irq > 0))
    request_irq(ctrl.csc_irq, &pcmcia_interrupt);
  
  if (register_chrdev(PCMCIA_MAJOR, "pc", &pcmcia_fops)) {
    printk("unable to get major %d for PCMCIA controller\n",
           PCMCIA_MAJOR);
    }
  
  return kmem_start;
  } /* pcmcia_init */

/*====================================================================*/

static void pcmcia_interrupt(int irq)
  {
  int i, ret, hit;
  if (pc_debug)
    printk("pcmcia_interrupt()\n");
  do {
    hit = 0;
    for (i = 0; i < slots_present; i++) {
      ret = ctrl.csc_poll(i);
      if (pc_debug)
	printk("  csc_poll(slot %d) = %d\n", i, ret);
      if (ret == 0) continue;
      hit++;
      if (slot_table[i].flags & SLOT_BUSY) {
        slot_table[i].flags |= SLOT_EVENT;
        slot_table[i].event |= ret;
        wake_up_interruptible(&slot_table[i].queue);
        }
      else
        printk("Bogus card status change on slot %d\n", i);
      }
    } while (hit > 0);
  } /* pcmcia_interrupt */
  
/*====================================================================*/

static int pcmcia_open(struct inode *inode, struct file *file)
  {
  u_short slot = MINOR(inode->i_rdev);
  if (pc_debug)
       printk("pcmcia_open(slot %d)\n", slot);
  if (slot >= slots_present) return -ENODEV;
  /* Only allow one writer at a time */
  if ((file->f_flags & O_ACCMODE) != O_RDONLY) {
    if (slot_table[slot].flags & SLOT_BUSY) return -EBUSY;
    slot_table[slot].flags |= SLOT_BUSY;
    ctrl.csc_on(slot, ctrl.csc_irq);
    }
  return 0;
  } /* pcmcia_open */

/*====================================================================*/

static void pcmcia_release(struct inode *inode, struct file *file)
  {
  u_short slot = MINOR(inode->i_rdev);
  if (pc_debug)
       printk("pcmcia_release(slot %d)\n", slot);
  if ((file->f_flags & O_ACCMODE) == O_RDONLY) return;
  slot_table[slot].flags &= ~SLOT_BUSY;
  if (slot_table[slot].flags & SLOT_EVENT) {
    slot_table[slot].flags &= ~SLOT_EVENT;
    slot_table[slot].event = 0;
    }
  ctrl.csc_off(slot);
  } /* pcmcia_release */

/*====================================================================*/

static int pcmcia_read(struct inode *inode, struct file *file,
                       char *buf, int count)
  {
  u_short slot = MINOR(inode->i_rdev);

  if (pc_debug)
       printk("pcmcia_read(slot %d)\n", slot);
  if (slot >= slots_present)
    return -ENODEV;
  if (count < 4)
    return -EINVAL;
  if ((file->f_flags & O_ACCMODE) == O_RDONLY) return -EBADF;
  if (ctrl.csc_irq) {
    if ((slot_table[slot].flags & SLOT_EVENT) == 0) {
      interruptible_sleep_on(&slot_table[slot].queue);
      if (current->signal & ~current->blocked) return -EINTR;
      }
    }
  else
    slot_table[slot].event = ctrl.csc_poll(slot);
  if (slot_table[slot].event == 0) return -EAGAIN;
  put_fs_long(slot_table[slot].event, buf);
  slot_table[slot].flags &= ~SLOT_EVENT;
  slot_table[slot].event = 0;
  return 4;
  } /* pcmcia_read */

/*====================================================================*/

static int pcmcia_select(struct inode *inode, struct file *file,
                         int sel_type, select_table *wait)
  {
  u_short slot = MINOR(inode->i_rdev);
  if (pc_debug)
    printk("pcmcia_select(slot %d)\n", slot);
  if (sel_type != SEL_IN)
    return 0;
  if ((file->f_flags & O_ACCMODE) == O_RDONLY) return -EBADF;
  if ((slot_table[slot].flags & SLOT_EVENT) || (ctrl.csc_irq == 0))
    return 1;
  select_wait(&slot_table[slot].queue, wait);
  return 0;
  } /* pcmcia_select */

/*====================================================================*/

static int pcmcia_ioctl(struct inode * inode, struct file * file,
			u_int cmd, u_long arg)
  {
  u_short slot = MINOR(inode->i_rdev);
  u_int size;
  int err;
  
  if (pc_debug)
    printk("pcmcia_ioctl(%d, %x, %lx)\n", slot, cmd, arg);
  if (slot >= slots_present) return -ENODEV;
  size = (cmd & IOCSIZE_MASK) >> IOCSIZE_SHIFT;
  if (cmd & IOC_IN) {
    /* check for *_SET_* ioctl calls */
    if (!(cmd & IOC_OUT)) {
      if ((file->f_flags & O_ACCMODE) == O_RDONLY)
	return -EBADF;
      if (!suser())
	return -EPERM;
      }
    err = verify_area(VERIFY_READ, (char *)arg, size);
    if (pc_debug)
      printk("i365_ioctl(%d, %x, %lx): verify_read = %d\n",
	     slot, cmd, arg, err);
    if (err) return err;
    }
  if (cmd & IOC_OUT) {
    err = verify_area(VERIFY_WRITE, (char *)arg, size);
    if (pc_debug)
      printk("i365_ioctl(%d, %x, %lx): verify_write = %d\n",
	     slot, cmd, arg, err);
    if (err) return err;
    }
  return ctrl.ioctl(slot, cmd, arg);
  } /* pcmcia_ioctl */
